import requests
import time

from .errors import WrongWorkflowProvided


class ComfyUIAPI:
    """
    This class makes connections between software and ComfyUI
    Allowing you to generate images remotely
    """
    _base_url = ""

    def __init__(self, comfyui_ip: str = '127.0.0.1', comfyui_port: str = '8188') -> None:
        """
        :param comfyui_ip str: IP address of ComfyUI web server.
        :param comfyui_port str: Port of ComfyUI web server.
        """
        self._base_url = f"http://{comfyui_ip}:{comfyui_port}"

    def send_generation_request(self, json_api_workflow: dict) -> dict:
        """
        Method that asks server to generate specific API Workflow.

        :param json_api_workflow dict: Dictonary object (json) containing workflow in API form
        :return dict: JSON answer from a server
        """
        if type(json_api_workflow) != dict:
            raise WrongWorkflowProvided(str(type(json_api_workflow)))
        return requests.post(f"{self._base_url}/prompt", json={'prompt': json_api_workflow}).json()

    def get_history(self, history_id: str) -> dict:
        """
        Method that asks server about specific generated image ID.
        Can return either empty dictionary if there is no such item in history
        or JSON API format else, containing everything from generation including result

        :param history_id str: ID of task, can be found in send_generation_request()['prompt_id']
        :return dict: JSON answer from a server
        """
        return requests.get(f"{self._base_url}/history/{history_id}").json()

    def download_item(self, path_to_save_file: str, comfy_filename: str, comfy_subfolder: str = '', comfy_type: str = 'output') -> None:
        """
        Method that allows to download a single image file.

        :param path_to_save_file str: Path where image would be saved
        :param comfy_filename str: File name inside ComfyUI output folder. Returns in get_history.
        :param comfy_subfolder str: Subfolder inside ComfyUI output folder. Returns in get_history.
        :param comfy_type str: Output type inside ComfyUI output folder. Returns in get_history.
        """
        with open(path_to_save_file, 'wb') as f:
            f.write(requests.get(f"{self._base_url}/view?filename={comfy_filename}&subfolder={comfy_subfolder}&type={comfy_type}").content)

    def generate_item(self, json_api_workflow: dict, save_path: str) -> None:
        """
        Full method that takes Workflow API and returns downloaded image.
        Exists for pure convinience.

        :param json_api_workflow dict: Dictonary object (json) containing workflow in API form
        :param save_path str: Path where image would be saved
        """
        # Step 1: Send generation request and record it's ID
        pID = self.send_generation_request(json_api_workflow)['prompt_id']

        # Step 2: Wait till history returns something (meaning, generation is finished)
        history = {}
        while len(history) == 0:
            time.sleep(2)
            history = self.get_history(pID)

        # Step 3: Save image somewhere
        for output in history[pID]['outputs']:
            self.download_item(save_path,
                               history[pID]['outputs'][output]['images'][0]['filename'],
                               history[pID]['outputs'][output]['images'][0]['subfolder'],
                               history[pID]['outputs'][output]['images'][0]['type'])
